package org.jicengine.element;

import org.jicengine.operation.StaticValue;
import org.jicengine.operation.OperationException;
import org.jicengine.operation.VariableValueOperation;
import org.jicengine.operation.Operation;
import org.jicengine.operation.EmptyOperation;
import org.jicengine.expression.LJEParser;
import org.jicengine.expression.ClassParser;
import org.jicengine.expression.SyntaxException;
import java.util.*;

/**
 * <p>
 * A class that makes it possible to create Element-objects little by little, as
 * is needed if the JIC-file is parsed with a SAX-parser.
 * </p>
 * <p>
 * ElementCompiler parses String-typed attributes and other data to corresponding
 * objects. After all the data of an XML-element has been processed, the resulting
 * runtime-Element can be obtained with method <code>createElement()</code>.
 * </p>
 * <p>
 *
 * </p>
 *
 * <h4> Element Life cycle </h4>
 *
 *
 * @author    .timo
 * @created   26. huhtikuuta 2004
 */

public abstract class ElementCompiler {

	private ElementImpl element;
	private String parameters;

	public ElementCompiler(String name, Location location)
	{
		this.element = new ElementImpl(name, location);
	}

	protected ElementImpl getElement()
	{
		return this.element;
	}

	public String getName()
	{
		return getElement().getName();
	}

	public Location getLocation()
	{
		return getElement().getLocation();
	}

	public void setOverridableBy(String overridingId)
	{
		getElement().setOverridableBy(overridingId);
	}

	public Element createElement() throws ElementException
	{
		return getElement().toRuntimeElement();
	}

	/**
	 * Sets the id of this element. Makes the value of this element available in
	 * the global-context
	 *
	 * @param id  The new id value
	 */
	public void setId(String id)
	{
		getElement().setId(id);
	}

	public void setParameters(String parameters)
	{
		this.parameters = parameters;
	}

	public void setConstructor(String expression) throws ElementException
	{
		try {
			getElement().setConstructor(LJEParser.getInstance().parse(expression));
		} catch (SyntaxException e){
			throw new ElementException("Failed to parse constructor '" + expression + "'.", e, getName(), getLocation());
		}
	}

	public Operation getConstructor()
	{
		return getElement().getConstructor();
	}

	public void setInstanceClass(String className) throws ElementException
	{
		try {
			getElement().setInstanceClass(ClassParser.INSTANCE.toClass(className));
		} catch(Exception e) {
			throw new ElementException("Failed to initialize attribute 'class'", e, getName(), getLocation());
		}
	}

	/**
	 * <p>
	 * for setting the action as an expression.
	 * </p>
	 * <p>
	 * enhancements:
	 * </p>
	 * <ul>
	 *  <li>empty string is mapped into an EmptyOperation</li>
	 *  <li>
	 *    expressions of type <code>setMethod(value)</code> and
	 *  <code>add(component)</code> are automatically madded to expressions
	 *  <code>PARENT.setMethod(value)</code> and
	 *  <code>PARENT.add(component)</code>.
	 *    </li>
	 * </ul>
	 */
	public void setAction(String expression) throws ElementException
	{
		if(expression.length() == 0) {
			getElement().setAction(EmptyOperation.INSTANCE);
		}
		else {
			String preparedExpression;
			int paramStart = expression.indexOf("(");
			int commaIndex = -1;

			if( paramStart != -1 ){
				commaIndex = expression.substring(0,paramStart).indexOf(".");
			}

			// the existence of '(' implies that the action is indeed a method call
			// (what else?)
			// the lack of '.' implies that the actor of the method call is missing.
			// therefore we add the implicit actor PARENT into the expression.
			if(paramStart != -1 && commaIndex == -1) {
				preparedExpression = Element.VARIABLE_NAME_PARENT_INSTANCE + "." + expression;
			}
			else {
				preparedExpression = expression;
			}

			try {
				getElement().setAction(LJEParser.getInstance().parse(preparedExpression));
			} catch (SyntaxException e){
				throw new ElementException("Failed to parse action.", e, getName(), getLocation());
			}
		}
	}


	/**
	 * if this element has a constructor, the cdata is added to the element-context
	 * with the name 'CDATA'. (cdata might be ignored if the constructor doesn't
	 * use CDATA) if there is no constructor, cdata becomes the value of this
	 * element.
	 *
	 * @param cData              Description of the Parameter
	 * @throws ElementException  Description of the Exception
	 */
	public void setCData(String cData) throws ElementException
	{
		if( getElement().getConstructor() == null ){
			// create an implicit constructor that returns the value of the cdata.
			getElement().setConstructor(new VariableValueOperation(Element.VARIABLE_NAME_CDATA));
		}

		// the CDATA is difficult to handle as a special case..
		// we create a virtual element that makes it possible to handle the
		// cdata like any other child element..
		ValueElement virtualElement = new StaticValueElement(Element.VARIABLE_NAME_CDATA, getLocation(), cData);

		// so .. then handle it like any other child element.
		handleChildElement(virtualElement);
	}

	/**
	 * Used for notifying this element about a child-element
	 *
	 * @param child              a child with no action, can have a value or not.
	 * @throws ElementException  Description of the Exception
	 */
	public void handleChildElement(Element child) throws ElementException
	{
		if( child instanceof ValueElement && !getElement().isNeeded((ValueElement)child) ){
			// hmm.. not a parameter. the child has a value but no action and therefore
			// no purpose. we have to do something with it.
			// we let the subclasses to decide the purpose of this child.
			ActionElement actionElement = handleExtraChildElement((ValueElement)child);

			// now the child is an action element and we can add it.
			getElement().addChildElement(actionElement);
		}
		else {
			getElement().addChildElement(child);
		}
	}

	/**
	 *
	 * for subclasses!
	 */
	protected abstract ActionElement handleExtraChildElement(ValueElement child) throws ElementException;


	/**
	 * Called by handler when the start-tag of the element has been processed -
	 * Element has been created and all attributes have been set, but no cdata nor
	 * child-elements have been processed. This is a good spot for verifying that
	 * the state of the element is valid - the attributes have been used properly,
	 * attributes not set by user can be set to their default values, etc.
	 *
	 * @throws ElementException  Description of the Exception
	 */
	public void elementInitialized() throws ElementException
	{
		// do some validity checks concerning the use of the 'parameters' attribute.
		if(this.parameters != null) {
			if(getElement().getInstanceClass() == null) {
				throw new ElementException("Attribute 'parameters' must be specified together with attribute 'class'", getName(), getLocation());
			}
			// NOTE: this checks works only if done before the implicit constructor is
			// created.
			if(getElement().getConstructor() != null) {
				// is it possible that the constructor in question is some default? if it is
				// should we just override it then?
				throw new ElementException("Attributes 'instance' and 'parameters' can't be used together", getName(), getLocation());
			}
		}

		createImplicitConstructor();
	}

	public String toString()
	{
		return "<" + getName() + ">";
	}

	/**
	 * <p>
	 * create the implicit constructor from 'class' and 'parameters' attributes,
	 * if possible. if the element already has a constructor, nothing is done.
	 * </p>
	 *
	 *
	 * @throws ElementException  Description of the Exception
	 */
	private void createImplicitConstructor() throws ElementException
	{
		if( getElement().getConstructor() == null && getElement().getInstanceClass() != null ){
			// we use the instanceClass-information for constructing the implicit constructor.

			String constructorExpression;
			if(this.parameters == null) {
				// empty constructor
				constructorExpression = "new " + getElement().getInstanceClass().getName() + "()";
			}
			else {
				// we have some parameters.
				constructorExpression = "new " + getElement().getInstanceClass().getName() + "(" + this.parameters + ")";
			}

			// TODO: don't create a string to be parsed - create a ready Operation-object

			try {
				setConstructor(constructorExpression);
			} catch (ElementException e){
				throw new ElementException("Problems creating implicit constructor.", e, getName(), getLocation());
			}
		}
	}

}
