package org.jicengine.element;

import org.jicengine.operation.SimpleContext;
import org.jicengine.operation.OperationException;
import org.jicengine.operation.LocalContext;
import org.jicengine.operation.Context;
import org.jicengine.operation.Operation;
import org.jicengine.BuildContext;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;

/**
 * <p>
 *
 * ElementImpl encapsulates the common properties and behaviour of an element.
 * </p> <h4> Common Element Structure </h4> <p>
 *
 * An Element consists of: </p>
 * <ul>
 *   <li> max 1 constructor together with * ValueElement-children as constructor
 *   parameters. </li>
 *   <li> * ActionElement-children </li>
 *   <li> max 1 action together with * ValueElement-children as action
 *   parameters. </li>
 * </ul>
 * <h4> Common processing procedures </h4>
 *
 * <p>
 * ElementImpl defines the common procedures for processing these
 * sub-components. The details of the processing can be tricky - an Element
 * might or might not have a constructor, it might or might not have an element,
 * etc.
 * </p>
 *
 * <h4> Lack of the exact runtime-behaviour </h4>
 * <p>
 * the runtime-behaviour of an Element is defined by interfaces ValueElement and
 * ActionElement. Whether an Element becomes an ValueElement or an ActionElement
 * depends on the state of the Element.
 * </p>
 * <p>
 *
 * ElementImpl can be initialized little-by-little, it is designed to be used
 * together with ElementCompiler. because of this approach, the
 * runtime-behaviour of the Element is known only when the Element is completely
 * initialized.
 * </p> <p>
 *
 * After an ElementImpl-object is completely initialized, its runtime-version
 * (ValueElement or ActionElement) can be obtained with method
 * toRuntimeElement(). </p>
 *
 * <h4> id binding </h4>
 * <p>
 * the object is added to the global-context only after it is created. this
 * way, the child-elements that are constructor parameters are processed before
 * the element has used its id-attribute. therefore, if a child has a same
 * id than the parent, the duplicate-id-exceptions will blame the wrong
 * element for the id - the parent.
 * </p>
 *
 * @author    .timo
 * @created   27. elokuuta 2004
 */

public class ElementImpl extends AbstractElement {

	/**
	 * ValueElement-version of an ElementImpl.
	 *
	 * @author    timo
	 * @created   27. elokuuta 2004
	 */
	public class ValueElementImpl implements ValueElement {
		public String getName()
		{
			return ElementImpl.this.getName();
		}

		public Object getValue(Context globalContext, Object parentInstance) throws ElementException
		{
			return ElementImpl.this.execute(globalContext, parentInstance);
		}
	}

	/**
	 * ActionElement-version of an ElementImpl.
	 *
	 * @author    timo
	 * @created   27. elokuuta 2004
	 */
	public class ActionElementImpl implements ActionElement {
		public String getName()
		{
			return ElementImpl.this.getName();
		}

		public void execute(Context globalContext, Object parentInstance) throws ElementException
		{
			ElementImpl.this.execute(globalContext, parentInstance);
		}
	}


	protected static final int CHILD_TYPE_CONSTRUCTOR_PARAM = 0;
	protected static final int CHILD_TYPE_ACTION_PARAM = 1;
	protected static final int CHILD_TYPE_ACTION = 2;

	private Operation constructor;
	private String id;
	private Class instanceClass;

	private String overridableBy;

	private List childElements = new ArrayList();
	private List childElementTypes = new ArrayList();
	private int lastConstructorParameterIndex = -1;

	private Operation action;

	public ElementImpl(String name, Location location)
	{
		super(name, location);
	}

	public void setConstructor(Operation operation)
	{
		if( this.constructor != null ){
			throw new ElementException("Constructor already set to '" + this.constructor + "', new constructor '" + operation + "' not accepted.",  getName(), getLocation());
		}
		this.constructor = operation;
	}

	public void setOverridableBy(String overridingId)
	{
		this.overridableBy = overridingId;
	}

	/**
	 * Sets the id of this element. Makes the value of this element available in
	 * the global-context
	 *
	 * @param id  The new id value
	 */
	public void setId(String id)
	{
		this.id = id;
	}

	public void setInstanceClass(Class instanceClass)
	{
		if( this.instanceClass != null ){
			throw new ElementException("Class already set to '" + this.instanceClass + "', new class '" + instanceClass + "' not accepted.",  getName(), getLocation());
		}
		this.instanceClass = instanceClass;
	}

	public void setAction(Operation operation) throws ElementException
	{
		if( this.action != null ){
			throw new ElementException("Action already set to '" + this.action + "', new action '" + operation + "' not accepted.",  getName(), getLocation());
		}
		this.action = operation;
	}

	public Operation getConstructor()
	{
		return this.constructor;
	}

	public Class getInstanceClass()
	{
		return this.instanceClass;
	}

	public Operation getAction()
	{
		return this.action;
	}

	protected boolean isOverridden(Context globalContext)
	{
		return (this.overridableBy != null && findBuildContextFrom(globalContext).getParameters().containsKey(this.overridableBy));
	}

	/**
	 * @param globalContext      the global context.
	 * @return                   null if this element can't be overridden at all or
	 *      is not overridden currently (overriding object not found).
	 * @throws ElementException  Description of the Exception
	 */
	private Object getOverridingObject(Context globalContext) throws ElementException
	{
		Object overridingObject = null;

		if(this.overridableBy != null) {
			// this element can be overridden..

			// obtain the build context
			BuildContext buildContext = findBuildContextFrom(globalContext);

			// obtain the parameters..
			Map buildParameters = buildContext.getParameters();

			if(buildParameters.containsKey(this.overridableBy)) {
				// this element is overridden!
				overridingObject = buildParameters.get(this.overridableBy);

				// validate it
				validateInstance(overridingObject, true);

			}
		}
		return overridingObject;
	}

	/**
	 * <p>
	 *
	 * Returns the runtime-version of this Element. Call this method after the
	 * Element is completely initialized. </p>
	 *
	 * @return                   either an instance of ValueElement or
	 *      ActionElement.
	 * @throws ElementException  Description of the Exception
	 */
	public Element toRuntimeElement() throws ElementException
	{
		// first a validity check:
		if(this.constructor == null && this.overridableBy == null) {
			// this element shouldn't have a value.

			if(this.instanceClass != null || this.id != null ) {
				// the existence of these parameters imply a value, however.
				throw new IllegalStateException("value element implied, but not enforced.");
			}
		}

		if(this.action != null) {
			return new ActionElementImpl();
		}
		else if(this.constructor != null || this.overridableBy != null) {
			return new ValueElementImpl();
		}
		else {
			// no action and no value?
			throw new ElementException("Element without action nor value is not allowed.", getName(), getLocation());
		}
	}

	public String toString()
	{
		return "<" + getName() + ">";
	}

	protected org.jicengine.BuildContext findBuildContextFrom(Context globalContext)
	{
		return (org.jicengine.BuildContext) globalContext.getObject(org.jicengine.BuildContext.VARIABLE_NAME);
	}

	/**
	 * @return true  if the child element is needed i.e. used by either the action
	 * or the constructor.
	 */
	public boolean isNeeded(ValueElement child)
	{
		return (this.constructor != null && this.constructor.needsParameter(child.getName())) || (this.action != null && this.action.needsParameter(child.getName()));
	}

	/**
	 * adds a child element into this element.
	 *
	 * @throws ElementException if the child is a ValueElement that is not needed
	 * by the action nor the constructor. unsused child elements without a purpose
	 * (without action) are not allowed because.. their value would be created
	 * and the result would be ignored.
	 */
	public void addChildElement(Element child) throws ElementException
	{
		if( child instanceof ActionElement ){
			// ok.
			addChildElement(child, CHILD_TYPE_ACTION);
		}
		else if( child instanceof ValueElement ){
			// a more problematic case.
			ValueElement valueChild = (ValueElement) child;

			// NOTE: constructor-param XOR action param!!
			if( getConstructor() != null && getConstructor().needsParameter(valueChild.getName()) ){
				// this is a constructor parameter.
				addChildElement(child, CHILD_TYPE_CONSTRUCTOR_PARAM);
			}
			else if( getAction() != null && getAction().needsParameter(valueChild.getName()) ){
				// this is a parameter of the action
				addChildElement(child, CHILD_TYPE_ACTION_PARAM);
			}
			else {
				// unused value element. we won't allow those.
				throw new ElementException("Unused child " + child + " without action not allowed.", getName(), getLocation());
			}
		}
		else {
			throw new ElementException("unknown child-element type: " + child, getName(), getLocation());
		}
	}

	private void addChildElement(Element child, int type)
	{
		this.childElements.add(child);
		this.childElementTypes.add(new Integer(type));
		if( type == CHILD_TYPE_CONSTRUCTOR_PARAM){
			// this is the currently last index.
			this.lastConstructorParameterIndex = this.childElements.size()-1;
		}
	}

	/**
	 * <p>
	 *
	 * returns the instance that is to be the value of this element. </p> <p>
	 *
	 * if this element is overridden, the returned instance will be the overrdigin
	 * object. </p> <p>
	 *
	 * if this element is not overridden, the instance will be created with the
	 * constructor. </p>
	 *
	 * @param globalContext      Description of the Parameter
	 * @return                   Description of the Return Value
	 * @throws ElementException  Description of the Exception
	 */
	private Object getInstance(Context constructorParams, Context globalContext) throws ElementException
	{

		Object instance;
		if(isOverridden(globalContext)) {
			instance = getOverridingObject(globalContext);
		}
		else if(this.constructor != null) {
			instance = createInstance(constructorParams, globalContext);
		}
		else if(this.overridableBy != null) {
			//
			throw new ElementException("Required overriding '" + this.overridableBy + "' not found.", getName(), getLocation());
		}
		else {
			throw new IllegalStateException("No constructor and no overriding - no instance available (" + getName() + " at " + getLocation() + ")");
		}

		if( this.id != null ){
			// we have to store the element into the context
			globalContext.addObject(this.id, instance);
		}

		return instance;
	}

	private Object executeConstructor(Context context) throws ElementException
	{
		try {
			//System.out.println("exec in '" + getName() + "' at " + getLocation());
			return this.constructor.execute(context);
		} catch(OperationException e) {
			throw new ElementException(e, getName(), getLocation());
		}
	}

	/**
	 * @param instance           the instance of this element, created by the
	 *      constructor. the instance might be needed in the action..
	 * @param globalContext      Description of the Parameter
	 * @throws ElementException  Description of the Exception
	 */
	private void executeAction(Context actionParameters, Context globalContext) throws ElementException
	{
		if(this.action == null) {
			throw new IllegalStateException("Action is null");
		}

		// context-hierarchy: action + global.
		Context actionContext = new LocalContext(
				actionParameters,
				globalContext);

		try {
			this.action.execute(actionContext);
		} catch(OperationException e) {
			// this way the location information is added to the exception.
			throw new ElementException("Failed to execute action " + this.action, e, getName(), getLocation());
		}
	}

	/**
	 * creates the instance (with the constructor), does the 'class'-attribute
	 * assertion, adds the instance to the global-context and performs the
	 * overriding, if necessary.
	 *
	 * @param globalContext           Description of the Parameter
	 * @return                        Description of the Return Value
	 * @throws ElementException       Description of the Exception
	 * @throws IllegalStateException  if this method was called although there is
	 *      no constructor.
	 */
	private Object createInstance(Context constructorParams, Context globalContext) throws ElementException
	{
		if(this.constructor == null) {
			throw new IllegalArgumentException("createInstance() called although constructor is null. (" + getName() + " at " + getLocation() + ")");
		}

		//create a context-hierarchy, where the constructor context objects
		// override the global objects.
		Context constructorContext = new LocalContext(
				constructorParams,
				globalContext);

		// now we can execute the constructor
		Object instance = executeConstructor(constructorContext);

		validateInstance(instance, false);

		return instance;
	}

	/**
	 * validates the instance. an exception is thrown, if the instance is not
	 * valid.
	 *
	 * currently, only the class of the instance is validated against the
	 * 'instanceClass'-property.
	 *
	 * @param isOverridingObject  for more better error messages..
	 *
	 * @throws Exception if the instance is not valid. since this method doesn't
	 * know where the instance came from, the caller must catch this exception
	 * and throw a new exception with a better error message.
	 */
	protected void validateInstance(Object instance, boolean isOverridingObject) throws ElementException
	{
		if( instance == null ){
			throw new ElementException("Instance is null - null values not allowed.", getName(), getLocation());
		}

		if( this.instanceClass != null && !this.instanceClass.isAssignableFrom(instance.getClass()) ){
			if( isOverridingObject ){
				throw new ElementException("The overriding object '" + this.overridableBy + "' was of type '" + instance.getClass().getName() + "', expected '" + this.instanceClass.getName() + "'", getName(), getLocation());
			}
			else {
				throw new ElementException("Expected the instance to be of type '" + this.instanceClass.getName() + "', was '" + instance.getClass().getName() + "'", getName(), getLocation());
			}
		}
	}

	/**
	 * <p>
	 * executes this element. this includes:
	 * </p>
	 * <ul>
	 *  <li> processing of all child elements, both value and action </li>
	 *  <li> executing the constructor (if there is one)</li>
	 *  <li> validating the instace (if there is one) </li>
	 *  <li> handling the overriding-issues: instance not created and only
	 *  action-parameter child-elements are executed if this element is overridden.</li>
	 *  <li> handling the id-attribute: the instance is added to the context, if
	 *  necessary.
	 *  <li> executing the action (if there is one) </li>
	 * </ul>
	 *
	 * @return  the instance of this element, if any, or null. the caller of
	 * this method should know whether the returned value is usable or not.
	 * if this element has an action, the returned value should not be used!!
	 * and note: the returned value won't be the result of the action. it will
	 * be the result of the constructor.
	 */
	protected Object execute(Context globalContext, Object parentInstance) throws ElementException
	{
		Context constructorParams = new SimpleContext();
		Context actionParams = new SimpleContext();

		if( parentInstance != null ){
			constructorParams.addObject(VARIABLE_NAME_PARENT_INSTANCE, parentInstance);
			actionParams.addObject(VARIABLE_NAME_PARENT_INSTANCE, parentInstance);
		}

		boolean isOverridden = isOverridden(globalContext);

		// assertion: lastConstructorParameterIndex <= childElements.size()!!

		// first set: child-elements that are executed before the instance is created.
		for (int i = 0; i <= this.lastConstructorParameterIndex; i++) {
			Element child = (Element) childElements.get(i);
			executeChild(child, ((Integer)childElementTypes.get(i)).intValue(), globalContext, null, constructorParams, actionParams, isOverridden);
		}

		Object instance = null;
		if( this.constructor != null || this.overridableBy != null) {
			// we have an instance.
			// obtain the instance - it is either created by the constructor or
			// is the overriding object.
			instance = getInstance(constructorParams, globalContext);

		}

		// execute rest of the children
		for (int i = (this.lastConstructorParameterIndex+1) ; i < childElements.size()  ; i++) {
			Element child = (Element) childElements.get(i);
			// the constructor params have no importance any more
			// should we make sure somehow that the constructorParams won't get 'abused'?
			executeChild(child, ((Integer)this.childElementTypes.get(i)).intValue(), globalContext, instance, constructorParams, actionParams, isOverridden);
		}

		if( this.action != null ){

			if( this.constructor != null || this.overridableBy != null ){
				// our instance could be needed in the action
				// note: we don't test 'if( instance != null)' because the support
				// for null-values could be implemented in the future.
				actionParams.addObject(Element.VARIABLE_NAME_ELEMENT_INSTANCE, instance);
			}

			executeAction(actionParams, globalContext);
		}

		// done.
		return instance;
	}

	private void executeChild(Element child, int type, Context globalContext, Object instance, Context constructorParams, Context actionParams, boolean isOverridden) throws ElementException
	{
		if( type == CHILD_TYPE_ACTION && !isOverridden){
			((ActionElement)child).execute(globalContext, instance);
		}
		else if( type == CHILD_TYPE_CONSTRUCTOR_PARAM && !isOverridden){
			Object value = ((ValueElement)child).getValue(globalContext, instance);
			constructorParams.addObject(child.getName(), value);
		}
		else if( type == CHILD_TYPE_ACTION_PARAM ){
			Object value = ((ValueElement)child).getValue(globalContext, instance);
			actionParams.addObject(child.getName(), value);
		}
		else {
			// ignored..
		}
	}
}
